# frozen_string_literal: true

module DrgDSL

  # Nodes in DRG logic AST.
  #
  # @see AstBuilder
  module Ast
    def self.node_classes
      Node.node_classes
    end

    module Node
      class << self
        def node_classes
          @node_classes ||= []
        end

        def included(node_class)
          node_classes << node_class
        end

        def type(node_class)
          node_class.to_s[/\w+$/]
        end
      end

      def accept(visitor)
        visitor.visit(self)
      end

      # Class name of the concrete node without the module name. Do not
      # override. Used by Visitor to determine which method to call.
      def type
        Node.type(self.class)
      end

      # @return [Boolean] whether this node represents a MDC equality check,
      #   e.g. "MDC = '01'"
      def mdc_equality?
        false
      end

      # @return [Boolean] whether this node represents a SEP equality check,
      #   e.g. "SEP = '07'"
      def sep_equality?
        false
      end

      # @return [Hash] Hash representation of the node
      def to_hash
        raise NotImplementedError
      end

      # @return [Integer] object hash for this AST node. Node instances for the
      #   same expression should always return the same hash code. This is
      #   useful for optimizations.
      def hash
        raise NotImplementedError
      end

      def ==(other)
        type == other.type && hash == other.hash
      end

      alias eql? ==

      def pretty_print(**pretty_printer_options)
        accept PrettyPrinter.new(**pretty_printer_options)
      end

      alias to_s pretty_print
    end

    # Represents a list of expressions joined with an "OR",
    # e.g., "exp1 or exp2 or exp3 ..."
    class Expression
      include Node

      attr_reader :expressions

      def initialize(expressions)
        @expressions = expressions
      end

      # @return [Boolean] whether this expression only contains SEP equality
      #   checks.
      def only_sep_equality_nodes?
        expressions.all?(&:sep_equality?)
      end

      # @return [Boolean] whether this expression only contains MDC equality
      #   checks.
      def only_mdc_equality_nodes?
        expressions.all?(&:mdc_equality?)
      end

      def to_hash
        { or: [expressions.map(&:to_hash)] }
      end

      def hash
        @hash ||= [type, *expressions].hash
      end
    end

    # Represents a list of expressions joined with an "AND",
    # e.g., "exp1 and exp2 and exp3 ..."
    class AndExpression
      include Node

      attr_reader :expressions

      def initialize(expressions)
        @expressions = expressions
      end

      def to_hash
        { and: [expressions.map(&:to_hash)] }
      end

      def hash
        @hash ||= [type, *expressions].hash
      end
    end

    # One expression surrounded by parentheses, e.g., "(exp)"
    class ParenExpression
      include Node

      attr_reader :expression

      def initialize(expression)
        @expression = expression
      end

      def to_hash
        { paren: expression.to_hash }
      end

      def hash
        @hash ||= [type, expression].hash
      end
    end

    # A NotExpression is a logic text of the form
    #   "not (expression)"
    class NotExpression
      include Node

      attr_reader :expression

      def initialize(expression)
        @expression = expression
      end

      def to_hash
        { not: expression.to_hash }
      end

      def hash
        @hash ||= [type, expression].hash
      end
    end

    # "Komplizierende_Prozeduren", "Dialyse", ...
    class FunctionCall
      include Node

      attr_reader :name

      def initialize(name)
        @name = name.to_s.strip
      end

      def to_hash
        { function_call: name }
      end

      def hash
        @hash ||= [type, name].hash
      end
    end

    # A drg link originates from the grammar rule basic_or_drg_link
    # This rule is of the form "variable (basic_expression | drg_link)"
    # therefore a DrgLink has a variable and a name
    class DrgLink
      include Node

      attr_reader :variable, :name

      # @param name [String] either a drg or adrg name
      def initialize(name:, variable:)
        @name = name.to_s.strip.upcase
        @variable = variable
      end

      def to_hash
        { drg_link: { var: variable.to_hash, name: name } }
      end

      def hash
        @hash ||= [type, name, variable].hash
      end
    end

    # A BasicExpression consists of a variable and a condition
    # (because every variant of condition has a variable)
    class BasicExpression
      include Node

      attr_reader :variable, :condition

      # @param variable [Variable]
      # @param condition [Comparison|UnaryCondition|TableCondition|Empty]
      def initialize(variable:, condition:)
        @variable = variable
        @condition = condition
      end

      def mdc_equality?
        variable.name == "MDC" && condition.op == "="
      end

      def sep_equality?
        variable.name == "SEP" && condition.op == "="
      end

      def to_hash
        { basic: { var: variable.to_hash, condition: condition.to_hash } }
      end

      def hash
        @hash ||= [type, variable, condition].hash
      end
    end

    # A comparison is either from a table_condition, a condition or a
    # date_expression
    class Comparison
      include Node

      attr_reader :op, :value, :table_condition

      def initialize(op:, value:, table_condition: nil)
        @op = op.to_s.strip
        @value = value
        @table_condition = table_condition
      end

      def to_hash
        {
          comparison: {
            op: op,
            value: value.to_hash,
            table_condition: table_condition&.to_hash
          }
        }
      end

      def hash
        @hash ||= [type, op, value, table_condition].hash
      end
    end

    # Has a certain operator (i.e. not / different) and a condition
    #
    # "not < 3"
    class UnaryCondition
      include Node

      attr_reader :op, :condition

      # @param op [String]
      def initialize(op:, condition:)
        @op = op.to_s.downcase
        @condition = condition
      end

      def to_hash
        {
          unary_condition: {
            op: op,
            condition: condition.to_hash
          }
        }
      end

      def hash
        @hash ||= [type, op, condition].hash
      end
    end

    class Empty
      include Node

      def to_hash
        { empty: nil }
      end

      def hash
        @hash ||= type.hash
      end
    end

    # There are 3 variants of table conditions (in_table, in_tables,
    # all_in_table)
    #
    # A table condition can have an optional comparison. Tables refers to the
    # table names (the internal names)
    class TableCondition
      IN_TABLE     = 'in table'
      IN_TABLES    = 'in tables'
      ALL_IN_TABLE = 'all in table'

      include Node

      attr_reader :op, :tables, :comparison

      # @param op [String] IN_TABLE, IN_TABLES or ALL_IN_TABLE
      # @param tables [Array<String>] one or more table names
      def initialize(op:, tables:, comparison: nil)
        @op = op
        @tables = Array(tables).map(&:to_s)
        @comparison = comparison
      end

      def to_hash
        {
          table_condition: {
            op: op,
            tables: tables,
            comparison: comparison&.to_hash
          }
        }
      end

      def hash
        @hash ||= [type, op, *tables, comparison].hash
      end
    end

    # INFO: srg (prozedur) l (links) r (rechts) b(beidseitig)
    class SrglrbTableCondition
      include Node

      attr_reader :variable, :condition

      def initialize(variable:, condition:)
        @variable = variable
        @condition = condition
      end

      def to_hash
        {
          srglrb_table_condition: {
            var: variable.to_hash,
            condition: condition.to_hash
          }
        }
      end

      def hash
        @hash ||= [type, variable, condition].hash
      end
    end

    class DateExpression
      include Node

      attr_reader :left_variable,
                  :right_variable,
                  :left_condition,
                  :right_condition,
                  :comparison,
                  :opd

      # @param opd [String] E.g. "opd4" => "mind. 4 erfüllen" query
      def initialize(left_variable:,
                     right_variable: nil,
                     left_condition:,
                     right_condition: nil,
                     comparison:,
                     opd:)

        @left_variable = left_variable
        @right_variable = right_variable
        @left_condition = left_condition
        @right_condition = right_condition
        @comparison = comparison
        @opd = opd.to_s.strip.upcase
      end

      def to_hash
        {
          opd: {
            var: opd,
            left_var: left_variable.to_hash,
            left_condition: left_condition.to_hash,
            right_var: right_variable&.to_hash,
            right_condition: right_condition&.to_hash,
            comparison: comparison.to_hash
          }
        }
      end

      def hash
        @hash ||= [
          type,
          left_variable,
          right_variable,
          left_condition,
          right_condition,
          comparison,
          opd
        ].hash
      end
    end

    # "PDX", "AGEYEARS", ...
    class Variable
      include Node

      attr_reader :name

      def initialize(name)
        @name = name.to_s.strip
      end

      def to_hash
        { name: name }
      end

      def hash
        @hash ||= [type, name].hash
      end
    end

    # "3", "'MDC'", ...
    class Constant
      include Node

      attr_reader :value

      def initialize(value)
        @value = value.to_s.strip.tr("'", '')
      end

      def to_s
        value.to_s
      end

      def to_hash
        { constant: { value: value } }
      end

      def hash
        @hash ||= [type, value].hash
      end
    end
  end
end
