# frozen_string_literal: true

require 'pp'

module DrgDSL
  class UnknownCstError < StandardError
    attr_reader :cst, :result

    # @param cst [Hash] CST obtained by parser
    # @param result [Object] whatever the AstBuilder was able to generate
    def initialize(cst, result)
      @cst = cst
      @result = result
    end

    def message
      <<~EOM
        Don't know how to build AST from this CST:

          #{pretty cst}

        Intermediate result:

          #{pretty result}
      EOM
    end

    private

    # @return [String] nicely formatted string for enhanced readability.
    def pretty(object)
      PP.pp object, ''
    end
  end

  class AstBuilder < Parslet::Transform

    # @param cst [Hash] CST-ish hash obtained from DrgParser.
    # @return [Node] AST
    # @raise [UnknownCstError] when CST could not be converted to a correct
    #   AST.
    def self.build(cst)
      ast = new.apply(cst)
      raise UnknownCstError.new(cst, ast) if broken_ast?(ast)
      ast
    end

    # Did we manage to build a correct AST?
    #
    # @return [Boolean]
    def self.broken_ast?(ast)
      !ast.is_a?(Ast::Node)
    end

    # Parslet transformation rules that transform a nested CST-ish hash
    # generated by DrgParser into a AST.
    #
    # Check out https://kschiess.github.io/parslet/transform.html for more
    # details on transformations.

    # Only one and-expression means there's no RHS of an "oder" expression, so
    # we just go downwards.
    rule(exp: subtree(:and_exp)) { and_exp }

    # Two or more and-expressions will be joined with an "oder". So and has
    # higher precedence.
    rule(exp: sequence(:and_exp)) do
      Ast::Expression.new(and_exp)
    end

    # In case of no RHS we simply go downwards.
    rule(and_exp: subtree(:simple)) do
      simple
    end

    rule(and_exp: sequence(:simple)) do
      Ast::AndExpression.new(simple)
    end

    rule(simple: subtree(:simple)) { simple }

    rule(paren_exp: subtree(:exp)) do
      Ast::ParenExpression.new(exp)
    end

    # "NOT ( ... )" -> "nicht [...]"
    rule(not_exp: subtree(:exp)) do
      Ast::NotExpression.new(exp)
    end

    # "EMPTY()" -> "leer"
    rule(empty: simple(:empty)) do
      Ast::Empty.new
    end

    # "NOT IN TABLE ... -> "nicht in Tabelle ...""
    rule(
      unary_condition: {
        keyword: simple(:keyword),
        condition: subtree(:condition)
      }
    ) do
      Ast::UnaryCondition.new(op: keyword, condition: condition)
    end

    rule(
      variable: subtree(:inner_variable)
    ) do
      inner_variable
    end

    # "AGEYEARS" --> "Alter"
    rule(inner_variable: subtree(:variable)) do
      Ast::Variable.new(variable)
    end

    # "42" -> "42"
    rule(constant: simple(:constant)) do
      Ast::Constant.new(constant)
    end

    # "DRG (F89)" -> "Definition der DRG (F98)"
    rule(
      basic_or_drg_link: {
        variable: subtree(:var),
        drg_link: {
          drg: simple(:drg)
        }
      }
    ) do
      Ast::DrgLink.new(name: drg, variable: var)
    end

    # "PDX IN TABLE ..." -> "Hauptdiagnose in Tabelle ..."
    rule(
      basic_or_drg_link: {
        variable: subtree(:var),
        condition: subtree(:condition)
      }
    ) do
      Ast::BasicExpression.new(variable: var, condition: condition)
    end

    # Go down to in-table, in-tables or all-in-tables condition.
    rule(table_condition: subtree(:table_condition)) { table_condition }

    # "in table (<reftable>)" -> "in Tabelle <external table name>"
    rule(in_table: simple(:table)) do
      Ast::TableCondition.new(
        op: Ast::TableCondition::IN_TABLE,
        tables: table
      )
    end

    # Multi-tables expression is deliberately mapped to one external table for
    # enhanced readability.
    #
    # "in tables (<ref1>, <ref2>, ...)" -> "in Tabelle <external table name>"
    rule(
      in_tables: {
        tables: sequence(:tables)
      }
    ) do
      Ast::TableCondition.new(
        op: Ast::TableCondition::IN_TABLES,
        tables: tables
      )
    end

    rule(
      comparison: subtree(:comparison)
    ) do
      comparison
    end

    rule(
      op: simple(:op),
      value: subtree(:value),
      table_condition: subtree(:table_condition)
    ) do
      Ast::Comparison.new(
        op: op,
        value: value,
        table_condition: table_condition
      )
    end

    # Handle dangling comparison after table expressions.
    rule(
      in_table: {
        table: simple(:table),
        comparison: subtree(:comparison)
      }
    ) do
      Ast::TableCondition.new(
        op: Ast::TableCondition::IN_TABLE,
        tables: table,
        comparison: comparison
      )
    end

    rule(
      in_tables: {
        tables: sequence(:tables),
        comparison: subtree(:comparison)
      }
    ) do
      Ast::TableCondition.new(
        op: Ast::TableCondition::IN_TABLES,
        tables: tables,
        comparison: comparison
      )
    end

    rule(table: simple(:table)) { table.to_s.strip }

    # "SRGLRB IN TABLE( CO3021ORS   )" -> "Beidseitige Prozedur in Tabelle CO3021ORS"
    rule(
      srglrb_table_condition: {
        variable: subtree(:var),
        condition: subtree(:condition)
      }
    ) do
      Ast::SrglrbTableCondition.new(variable: var, condition: condition)
    end

    # "OPD2 in (SRG in table ...) > 0" -> "Mindestens zwei Behandlungen, die mindestens einen Tag auseinander liegen in [Prozedur in Tabelle ...]" ¯\_(ツ)_/¯
    rule(
      date_exp: {
        opd: simple(:opd),
        variable: subtree(:left_var),
        left_table_condition: subtree(:left_table_condition),
        and_table_condition: {
          variable: subtree(:right_var),
          right_table_condition: subtree(:right_table_condition)
        },
        comparison: subtree(:comparison)
      }
    ) do
      Ast::DateExpression.new(
        left_variable: left_var,
        right_variable: right_var,
        left_condition: left_table_condition,
        right_condition: right_table_condition,
        comparison: comparison,
        opd: opd
      )
    end

    rule(
      date_exp: {
        opd: simple(:opd),
        variable: subtree(:left_var),
        left_table_condition: subtree(:left_table_condition),
        comparison: subtree(:comparison)
      }
    ) do
      Ast::DateExpression.new(
        left_variable: left_var,
        left_condition: left_table_condition,
        comparison: comparison,
        opd: opd
      )
    end

    # "Vierzeitige_bestimmte_OR" -> "Vierzeitige bestimmte OR-Prozeduren"
    rule(function_call: simple(:fname)) do
      Ast::FunctionCall.new(fname)
    end
  end
end
