# frozen_string_literal: true

module DrgDSL
  class ParserError < StandardError
    attr_reader :parslet_error, :input

    def initialize(parslet_error:, input:)
      @parslet_error = parslet_error
      @input = input
    end

    def message
      %{Failed to parse "#{input}"\nParslet::ParseFailed: #{parslet_error}}
    end
  end

  # Used to parse logic expressions in flowchart decision tree. Based on the
  # Grouper's DRG parser[0], only difference being that this parser doesn't
  # know about statements, since they do not occur in the decision nodes.
  #
  # To learn more about Parslet see the parser[1] and transform[2] docs.
  #
  # The full syntax of the logic is described in
  # documents/Spec-Handbuch_v2.2.3.pdf in chapter 5.5.
  #
  # [0] https://github.com/swissdrg/grouper/blob/master/src/main/common/org/swissdrg/grouper/parser/DrgParser.jj
  # [1] https://kschiess.github.io/parslet/parser.html
  # [2] https://kschiess.github.io/parslet/transform.html
  class Parser < Parslet::Parser

    # @param expression [String]
    # @return [Ast::Node]
    def self.parse(expression)
      AstBuilder.build new.parse(expression.to_s.strip)
    rescue Parslet::ParseFailed => e
      raise ParserError.new(parslet_error: e, input: expression)
    end

    # Case-insensitive string matching
    #
    # https://kschiess.github.io/parslet/tricks.html
    def stri(str)
      key_chars = str.split(//)
      key_chars
        .map { |c| match["#{c.upcase}#{c.downcase}".squeeze] }
        .inject(:>>)
    end

    # Tokens
    rule(:star)          { str('*') >> sp? }
    rule(:k_or)          { stri('or') }
    rule(:k_and)         { stri('and') }
    rule(:k_not)         { stri('not') }
    rule(:k_different)   { stri('different') }
    rule(:k_empty)       { stri('empty') }
    rule(:k_in)          { stri('in') }
    rule(:mdc)           { stri('mdc') }
    rule(:in_table)      { stri('in') >> sp? >> stri('table') }
    rule(:in_tables)     { stri('in') >> sp? >> stri('tables') }
    rule(:all_in_table)  { stri('all') >> sp? >> stri('in') >> sp? >> stri('table') }
    rule(:lpar)          { str('(') >> sp? }
    rule(:rpar)          { sp? >> str(')') }
    rule(:underscore)    { str('_') }
    rule(:comma)         { str(',') }
    rule(:quote)         { str("'") }
    rule(:assign)        { str(':=') }
    rule(:letter)        { match('[a-zA-Z]') }
    rule(:digit)         { match('[0-9]') }
    rule(:number)        { digit.repeat(1) }

    rule(:comparison_operator) do
      str(">=")   |
        str(">")  |
        str("<=") |
        str("<")  |
        str("=")
    end

    #   name
    #     ::= letter (letter | number | underscore)*
    rule(:name)          { letter >> (letter | number | underscore).repeat }

    rule(:variable_name) do
      stri("PDX") |
        stri("SDX") >> number.maybe |
        stri("DDX") >> number.maybe |
        stri("SRG") >> number.maybe |
        stri("SRGB") >> number.maybe |
        stri("SRGR") >> number.maybe |
        stri("SRGL") >> number.maybe |
        stri("SRGN") >> number.maybe |
        stri("SRGU") >> number.maybe |
        stri("AGEYEARS") |
        stri("AGEDAYS") |
        stri("SEX") |
        stri("ADM_WT") |
        stri("HMV") |
        stri("MDC") |
        stri("DRG") |
        stri("ADRG") |
        stri("GST") |
        stri("LOS") |
        stri("PCCL") |
        stri("SEP") |
        stri("ENTRY") |
        stri("ADM_MODE") |
        stri("VOID") |
        stri("GESTAGE")
    end

    #   srglrb
    #     ::= 'SRGLRB' number{0,1}
    rule(:srglrb) { stri('SRGLRB') >> number.maybe }

    #   opd
    #     ::= 'OPD' number{0,1}
    rule(:opd) { stri('OPD') >> number.maybe }

    # By "double nesting" the variable, the AST builder (a parslet transformer)
    # can use subtree(:variable) to retrieve a variable node, given there's a
    # rule for inner_variable that creates said node.
    #
    #   variable
    #     ::= variable_name
    rule(:variable) { (star.maybe >> variable_name).as(:inner_variable).as(:variable) }

    #   constant
    #     ::= quote{0,1} (number | name) name{0,1} quote{0,1}
    rule(:constant) do
      (quote.maybe >> ((number | name) >> name.maybe) >> quote.maybe).as(:constant)
    end

    #   value
    #     ::= variable | constant
    rule(:value) { (variable | constant).as(:value) }

    #   function_call
    #     ::= name
    rule(:function_call) do
      name.as(:function_call)
    end

    #   empty
    #     ::= empty '()'
    rule(:empty) do
      k_empty >> sp? >> (lpar >> rpar).maybe
    end

    #   table_condition
    #     ::= (in_table '(' name ')' comparison{0,1})
    #       | in_tables '(' name (comma name)* ')' comparison{0,1}
    #       | all_in_table '(' name ')'
    rule(:table_condition) do
      (in_table >> sp? >> lpar >> name.as(:table) >> rpar >> sp? >> comparison.as(:comparison).maybe).as(:in_table) |
        (in_tables >> sp? >> lpar >> (name.as(:table) >> (sp? >> comma >> sp? >> name.as(:table)).repeat).as(:tables) >> rpar >> sp? >> comparison.as(:comparison).maybe).as(:in_tables) |
        (all_in_table >> sp? >> lpar >> name.as(:table) >> rpar).as(:all_in_table)
    end

    #   comparison
    #     ::= comparison_operator value table_condition{0,1}
    rule(:comparison) do
      comparison_operator.as(:op) >> sp? >>
        value >> sp? >> table_condition.maybe.as(:table_condition)
    end

    #   unary_condition
    #     ::= (not | different) condition
    rule(:unary_condition) do
      (k_not | k_different).as(:keyword) >> sp? >> condition.as(:condition)
    end

    #   condition
    #     ::= comparison
    #       | unary_condition
    #       | empty
    #       | table_condition
    rule(:condition) do
      comparison.as(:comparison) | unary_condition.as(:unary_condition) | empty.as(:empty) | table_condition.as(:table_condition)
    end

    #   expression
    #     ::= and_expression (or and_expression)*
    rule(:expression) do
      (sp? >> and_expression.as(:and_exp) >> (sp? >> k_or >> sp? >> and_expression.as(:and_exp)).repeat >> sp?).as(:exp)
    end

    #   and_expression
    #     ::= simple_expression (and simple_expression)*
    rule(:and_expression) do
      simple_expression.as(:simple) >> (sp? >> k_and >> sp? >> simple_expression.as(:simple)).repeat
    end

    #   simple_expression
    #     ::= '(' expression ')'
    #       | basic_or_drg_link
    #       | srglrb_table_condition
    #       | date_expression
    #       | not_expression
    #       | function_call
    rule(:simple_expression) do
      (lpar >> expression >> rpar).as(:paren_exp) |
        basic_or_drg_link |
        srglrb_table_condition |
        date_expression |
        not_expression |
        function_call
    end

    #   date_expression
    #     ::= opd in '(' variable table_condition (and variable table_ondition){0,1} ')' comparison
    rule(:date_expression) do
      (opd.as(:opd) >> sp? >> k_in >> sp? >> lpar >> variable >> sp? >> table_condition.as(:left_table_condition) >> (sp? >> k_and >> sp? >> variable >> sp? >> table_condition.as(:right_table_condition)).as(:and_table_condition).maybe >> rpar >> sp? >> comparison.as(:comparison)).as(:date_exp)
    end

    #   srglrb_table_condition
    #     ::= star{0,1} srglrb condition
    rule(:srglrb_table_condition) do
      ((star.maybe >> srglrb).as(:inner_variable).as(:variable) >> sp? >> condition.as(:condition)).as(:srglrb_table_condition)
    end

    #   not_expression
    #     ::= not '(' expression ')'
    rule(:not_expression) do
      (k_not >> sp? >> lpar >> expression >> rpar).as(:not_exp)
    end

    #   basic_or_drg_link
    #     ::= variable (basic_expression | drg_link)
    rule(:basic_or_drg_link) do
      (variable >> sp? >> (basic_expression | drg_link)).as(:basic_or_drg_link)
    end

    #   basic_expression
    #     ::= condition
    rule(:basic_expression) do
      condition.as(:condition)
    end

    #   drg_link
    #     ::= '(' name ')'
    rule(:drg_link) do
      (lpar >> name.as(:drg) >> rpar).as(:drg_link)
    end

    rule(:comment) do
      str('/*') >> (str('*/').absent? >> any).repeat >> str('*/')
    end

    rule(:sp) do
      (match("[ \t\r\n\f]") | comment).repeat(1)
    end

    # Optional whitespace rule. Unfortunately, skipping whitespace can't be
    # automated:
    # https://github.com/kschiess/parslet/issues/49
    rule(:sp?) { sp.repeat }

    rule(:root) { expression }
  end
end
