# frozen_string_literal: true

require 'cgi'

module DrgDSL
  class PrettyPrinter
    using CoreExtensions
    include Visitor

    OUTPUT_FORMATS = %i[string html bash].freeze
    OUTPUT_FORMATS.each do |format|
      define_method "output_#{format}?" do
        @output_format == format
      end
    end

    attr_reader :indentation_level, :indentation_string, :output_format

    # @param expression [String]
    # @return [String]
    def self.pretty_print(expression, **options)
      ast = Parser.parse(expression)
      ast = ParenCleaner.clean(ast) if options.delete(:remove_redundant_parens)

      output = ast.accept(new(**options))
      output_format = options[:output_format]
      return "<pre>#{output}</pre>" if output_format == :html
      output
    end

    # @param output_format [Symbol]
    # @param multiline [Boolean]
    def initialize(output_format: :string,
                   multiline: true)

      @indentation_level = 0
      @indentation_string = '  '

      unless OUTPUT_FORMATS.include?(output_format)
        raise "Unknown output format: #{output_format.inspect}"
      end

      @output_format = output_format
      @multiline = multiline
    end

    def multiline?
      @multiline
    end

    def visit_nil
      ''
    end

    def cache?
      true
    end

    def cache_key(n)
      [
        n.hash,
        @indentation_level,
        @indentation_string,
        @output_format,
        @multiline
      ].hash
    end

    private

    def visit_Expression(n)
      keyword = wrap('OR', type: :keyword)

      separator = " #{keyword} "
      separator = "\n#{f(keyword)} " if multiline?

      n.expressions.map { |e| visit(e) }.join(separator)
    end

    def visit_AndExpression(n)
      keyword = wrap('AND', type: :keyword)

      separator = " #{keyword} "
      separator = "\n#{f(keyword)} " if multiline?

      n.expressions.map { |e| visit(e) }.join(separator)
    end

    def visit_ParenExpression(n)
      return "(#{visit(n.expression)})" unless multiline?

      str = +"(\n"
      indented do
        str << f(visit(n.expression))
      end
      str << "\n"
      str << f(')')
    end

    def visit_NotExpression(n)
      "#{wrap('NOT', type: :keyword)} (#{visit(n.expression)})"
    end

    def visit_FunctionCall(n)
      wrap(n.name, type: :function)
    end

    def visit_DrgLink(n)
      "#{visit(n.variable)} (#{wrap(n.name, type: n.variable.name.downcase)})"
    end

    def visit_BasicExpression(n)
      "#{visit(n.variable)} #{visit(n.condition)}"
    end

    def visit_Comparison(n)
      "#{wrap(n.op, type: :op)} #{visit(n.value)} #{visit(n.table_condition)}".strip
    end

    def visit_UnaryCondition(n)
      "#{wrap(n.op.upcase, type: :keyword)} #{visit(n.condition)}"
    end

    def visit_Empty(_)
      wrap('EMPTY', type: :keyword)
    end

    def visit_TableCondition(n)
      str = +"#{wrap(n.op.upcase, type: :keyword)} "

      if multiline? && n.tables.count > 4
        str << "(\n"
        indented do
          str << n.tables.map { |t| f(wrap(t.upcase, type: :table)) }.join(",\n")
        end
        str << "\n"
        str << f(')')
      else
        str << "(#{n.tables.map { |t| wrap(t.upcase, type: :table) }.join(', ')})"
      end

      str
    end

    def visit_SrglrbTableCondition(n)
      "#{visit(n.variable)} #{visit(n.condition)}"
    end

    def visit_DateExpression(n)
      "#{wrap(n.opd, type: :variable)} IN (#{visit(n.left_variable)} #{visit(n.left_condition)}#{n.right_variable && " AND #{visit(n.right_variable)} #{visit(n.right_condition)}"}) #{visit(n.comparison)}".strip
    end

    def visit_Variable(n)
      wrap(n.name.upcase, type: :variable)
    end

    def visit_Constant(n)
      wrap(n.value, type: :constant)
    end

    def indented
      indent!
      yield
      deindent!
    end

    def indent!
      @indentation_level += 1
    end

    def deindent!
      @indentation_level -= 1
    end

    def indentation
      return '' unless multiline?
      @indentation_string * @indentation_level
    end

    # @param str [String]
    # @return [String] indented string
    def f(str)
      "#{indentation}#{str}"
    end

    def wrap(str, **options)
      return str if output_string?
      return span(str, type: options[:type]) if output_html?
      return ansi_colored(str, type: options[:type]) if output_bash?
    end

    def span(str, type:)
      class_name = "drgdsl-#{type}"
      %{<span class="#{class_name}">#{escape_html(str)}</span>}
    end

    def escape_html(str)
      CGI.escapeHTML(str)
    end

    def ansi_colored(str, type:)
      case type.to_sym
      when :keyword
        str.blue
      when :function
        str.green
      when :op
        str.blue
      when :table
        str.magenta
      when :variable
        str.cyan
      when :constant
        str.red
      when :drg, :adrg
        str.yellow
      else
        str
      end
    end
  end
end
