# frozen_string_literal: true

module DrgDSL

  # Mixin for Ast::Node visitors. Visits all child nodes and returns the
  # visited node by default.
  module Visitor
    def self.included(visitor)
      visitor.prepend Cache
      visitor.extend ClassMethods
    end

    module Cache
      # @param n [Ast::Node]
      def visit(n)
        return super(n) unless cache?
        self.class.cache[cache_key(n)] ||= super(n)
      end
    end

    module ClassMethods
      # @return [Hash] in-memory Ruby Hash
      def cache
        @cache ||= {}
      end

      def clear_cache!
        @cache = {}
      end
    end

    # Whether this visitor should cache visit results. Note that one might need
    # to override #cache_key(n) if a visit is not deterministic, e.g. when it
    # depends on some instance variables.
    #
    # @see PrettyPrinter#cache_key
    #
    # @return [Boolean]
    def cache?
      false
    end

    # @param n [Ast::Node]
    # @return [Integer] ideally an object hash
    def cache_key(n)
      n.hash
    end

    # @param n [Ast::Node]
    def visit(n)
      return visit_nil if n.nil?
      send("visit_#{n.type}", n)
    end

    # Override to define what should happen when one calls #visit with nil (can
    # occur on optional AST branches).
    #
    # Does nothing by default. A string visitor might want to return an empty
    # string here.
    def visit_nil
    end

    def visit_Expression(n)
      n.expressions.map { |e| visit(e) }
      n
    end

    def visit_AndExpression(n)
      n.expressions.map { |e| visit(e) }
      n
    end

    def visit_ParenExpression(n)
      visit(n.expression)
      n
    end

    def visit_NotExpression(n)
      visit(n.expression)
      n
    end

    def visit_FunctionCall(n)
      n
    end

    def visit_DrgLink(n)
      visit(n.variable)
      n
    end

    def visit_BasicExpression(n)
      visit(n.variable)
      visit(n.condition)
      n
    end

    def visit_Comparison(n)
      visit(n.value)
      visit(n.table_condition)
      n
    end

    def visit_UnaryCondition(n)
      visit(n.condition)
      n
    end

    def visit_Empty(n)
      n
    end

    def visit_TableCondition(n)
      visit(n.comparison)
      n
    end

    def visit_SrglrbTableCondition(n)
      visit(n.condition)
      n
    end

    def visit_DateExpression(n)
      visit(n.left_variable)
      visit(n.left_condition)
      visit(n.right_variable)
      visit(n.right_condition)
      visit(n.comparison)
      n
    end

    def visit_Variable(n)
      n
    end

    def visit_Constant(n)
      n
    end
  end
end
